import sys
import functools
import inspect
from types import TracebackType
from typing import Any, Callable, cast, TypeVar


T = TypeVar("T", bound=Callable[..., Any])


def streamline_tracebacks() -> Callable[[T], T]:
    def decorator(func: T) -> T:
        is_async = inspect.iscoroutinefunction(func)

        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            try:
                return func(*args, **kwargs)
            except Exception as e:
                raise e.with_traceback(streamlined_traceback())

        @functools.wraps(func)
        async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
            try:
                return await func(*args, **kwargs)
            except Exception as e:
                raise e.with_traceback(streamlined_traceback())

        return cast(T, async_wrapper if is_async else wrapper)

    return decorator


def streamlined_traceback() -> TracebackType | None:
    traceback = sys.exc_info()[2]
    if traceback is None:
        return None
    child_traceback, _ = get_child_traceback(traceback)
    return child_traceback


def get_child_traceback(
    traceback: TracebackType,
) -> tuple[TracebackType, bool]:
    if not traceback.tb_next:
        return traceback, False
    next_traceback, mp_actor_code = get_child_traceback(traceback.tb_next)
    if mp_actor_code:
        return next_traceback, True
    elif "/mp_actors/" in traceback.tb_frame.f_code.co_filename:
        return next_traceback, True
    return traceback, False



import time
import asyncio
import functools
import logging
import inspect
from typing import (
    Any,
    Callable,
    Coroutine,
    TypeVar,
    Optional,
    Union,
)

T = TypeVar("T")
R = TypeVar("R")


def retry(
    max_attempts: int = 3,
    delay: float = 0.25,
    backoff_factor: float = 2.0,
    exceptions: tuple = (Exception,),
    on_retry: Optional[Callable[[Exception, int], None]] = None,
) -> Callable[
    [Callable[..., Union[T, Coroutine[Any, Any, R]]]],
    Callable[..., Union[T, Coroutine[Any, Any, R]]],
]:
    """
    Retry decorator with exponential backoff for both synchronous and asynchronous functions.

    Args:
        max_attempts: Maximum number of retry attempts
        delay: Initial delay between retries in seconds
        backoff_factor: Multiplier for delay between retries
        exceptions: Tuple of exceptions to catch and retry on
        on_retry: Optional callback function called on retry with the exception and attempt number

    Returns:
        Decorated function that will retry on specified exceptions
    """

    def decorator(func):
        is_coroutine = inspect.iscoroutinefunction(func)

        @functools.wraps(func)
        def sync_wrapper(*args: Any, **kwargs: Any) -> T:
            last_exception = None
            current_delay = delay

            for attempt in range(1, max_attempts + 1):
                try:
                    return func(*args, **kwargs)
                except exceptions as e:
                    last_exception = e

                    if attempt == max_attempts:
                        raise

                    if on_retry:
                        on_retry(e, attempt)
                    else:
                        logging.warning(
                            f"Retry {attempt}/{max_attempts} for {func.__name__} "
                            f"after error: {str(e)}"
                        )

                    time.sleep(current_delay)
                    current_delay *= backoff_factor

            assert last_exception is not None
            raise last_exception

        @functools.wraps(func)
        async def async_wrapper(*args: Any, **kwargs: Any) -> R:
            last_exception = None
            current_delay = delay

            for attempt in range(1, max_attempts + 1):
                try:
                    return await func(*args, **kwargs)
                except exceptions as e:
                    last_exception = e

                    if attempt == max_attempts:
                        raise

                    if on_retry:
                        on_retry(e, attempt)
                    else:
                        logging.warning(
                            f"Retry {attempt}/{max_attempts} for {func.__name__} "
                            f"after error: {str(e)}"
                        )

                    await asyncio.sleep(current_delay)
                    current_delay *= backoff_factor

            assert last_exception is not None
            raise last_exception

        return async_wrapper if is_coroutine else sync_wrapper

    return decorator



import asyncio
import pydantic
import traceback
from typing import Awaitable, cast, Iterable, Iterator, overload

from .types import MessagesAndChoices


MetadataValue = float | int | str | bool | None


class PydanticException(pydantic.BaseModel):
    type: str
    message: str
    traceback: str


class Trajectory(pydantic.BaseModel):
    messages_and_choices: MessagesAndChoices
    reward: float
    metrics: dict[str, float] = {}
    metadata: dict[str, MetadataValue] = {}


class TrajectoryGroup(pydantic.BaseModel):
    trajectories: list[Trajectory]
    metadata: dict[str, MetadataValue] = {}
    exceptions: list[PydanticException] = []

    def __init__(
        self,
        trajectories: Iterable[Trajectory] | Iterable[Awaitable[Trajectory]],
        *,
        metadata: dict[str, MetadataValue] = {},
        exceptions: list[BaseException] = [],
    ) -> None:
        super().__init__(
            trajectories=list(trajectories),
            metadata=metadata,
            exceptions=[
                PydanticException(
                    type=str(type(e)),
                    message=str(e),
                    traceback="\n".join(
                        traceback.format_exception(type(e), e, e.__traceback__)
                    ),
                )
                for e in exceptions
            ],
        )

    def __iter__(self) -> Iterator[Trajectory]:
        return iter(self.trajectories)

    def __len__(self) -> int:
        return len(self.trajectories)

    @overload
    def __new__(
        cls,
        trajectories: Iterable[Trajectory],
        *,
        metadata: dict[str, MetadataValue] = {},
        exceptions: list[BaseException] = [],
    ) -> "TrajectoryGroup": ...

    @overload
    def __new__(
        cls,
        trajectories: Iterable[Awaitable[Trajectory]],
        *,
        metadata: dict[str, MetadataValue] = {},
        exceptions: list[BaseException] = [],
    ) -> Awaitable["TrajectoryGroup"]: ...

    def __new__(
        cls,
        trajectories: Iterable[Trajectory] | Iterable[Awaitable[Trajectory]],
        *,
        metadata: dict[str, MetadataValue] = {},
        exceptions: list[BaseException] = [],
    ) -> "TrajectoryGroup | Awaitable[TrajectoryGroup]":
        ts = list(trajectories)
        if all(isinstance(t, Trajectory) for t in ts):
            group = super().__new__(cls)
            group.__init__(
                trajectories=cast(list[Trajectory], ts),
                metadata=metadata,
                exceptions=exceptions,
            )
            return group
        else:

            async def _(exceptions: list[BaseException]):
                from .gather import get_gather_context, record_metrics

                context = get_gather_context()
                trajectories = []
                for future in asyncio.as_completed(
                    cast(list[Awaitable[Trajectory]], ts)
                ):
                    try:
                        trajectory = await future
                        trajectories.append(trajectory)
                        record_metrics(context, trajectory)
                        context.update_pbar(n=1)
                    except BaseException as e:
                        exceptions.append(e)
                        context.metric_sums["exceptions"] += 1
                        context.update_pbar(n=0)
                        if context.too_many_exceptions():
                            raise
                return TrajectoryGroup(
                    trajectories=trajectories,
                    exceptions=exceptions,
                    metadata=metadata,
                )

            class CoroutineWithMetadata:
                def __init__(self, coro, num_trajectories):
                    self.coro = coro
                    self._num_trajectories = num_trajectories

                def __await__(self):
                    return self.coro.__await__()

            coro = _(exceptions.copy())
            return CoroutineWithMetadata(coro, len(ts))



from typing import Callable, Coroutine, Any

import art

from ..trajectories import Trajectory, TrajectoryGroup


async def benchmark_rollout(
    model: str,
    num_rollouts: int,
    rollout: Callable[[str, int, bool], Coroutine[Any, Any, Trajectory]],
) -> float:
    trajectory_groups = await art.gather_trajectory_groups(
        [TrajectoryGroup(rollout(model, i, False) for i in range(num_rollouts))],
        pbar_desc="Benchmarking rollout",
    )

    trajectory_group_rewards = []

    for group in trajectory_groups:
        total_reward = sum(trajectory.reward for trajectory in group)
        trajectory_group_rewards.append(total_reward / len(group))

    average_reward = sum(trajectory_group_rewards) / len(trajectory_group_rewards)

    print(f"Average reward for {model}: {average_reward}")

    return average_reward



from ..trajectories import Trajectory
from ..types import Messages


def get_trajectory_messages(trajectory: Trajectory) -> Messages:
    messages: Messages = []
    for item in trajectory.messages_and_choices:

        # if item is not a dict, convert it to a dict
        if not isinstance(item, dict):
            item = item.to_dict()

        # check if item is a choice
        if "message" in item:
            messages.append(
                {"role": "assistant", "content": item["message"]["content"]}
            )
        else:
            # otherwise it's a message
            messages.append(item)
    return messages



import art
import json
import random
from typing import TypedDict
from typing import Literal
from pydantic import BaseModel


class TicTacToeGame(TypedDict):
    board: list[list[str]]
    agent_symbol: Literal["x", "o"]
    opponent_symbol: Literal["x", "o"]


def generate_game(board_length: int = 3) -> TicTacToeGame:
    board = [["_" for _ in range(board_length)] for _ in range(board_length)]
    agent_symbol = random.choice(["x", "o"])
    opponent_symbol = "x" if agent_symbol == "o" else "o"
    return {
        "board": board,
        "agent_symbol": agent_symbol,
        "opponent_symbol": opponent_symbol,
    }


def render_board(game: TicTacToeGame) -> str:
    board = game["board"]
    board_length = len(board)
    # print something like this:
    #    1   2   3
    # A  _ | x | x
    # B  o | _ | _
    # C  _ | o | _
    # where _ is an empty cell

    board_str = "   " + "   ".join([str(i + 1) for i in range(board_length)]) + "\n"
    for i in range(board_length):
        board_str += f"{chr(65 + i)}  {board[i][0]} | {board[i][1]} | {board[i][2]}\n"
    return board_str


def get_opponent_move(game: TicTacToeGame) -> tuple[int, int]:
    # get a random empty cell
    empty_cells = [
        (i, j) for i in range(3) for j in range(3) if game["board"][i][j] == "_"
    ]
    return random.choice(empty_cells)


class AgentMove(BaseModel):
    reason: str
    square: str


def apply_agent_move(game: TicTacToeGame, move: str) -> None:
    board_length = len(game["board"])
    json_move = json.loads(move)
    square = json_move["square"]

    try:
        row_index = ord(square[0]) - 65
        col_index = int(square[1]) - 1
    except Exception as e:
        print(e)
        raise ValueError("Unable to parse square")

    if (
        row_index < 0
        or row_index >= board_length
        or col_index < 0
        or col_index >= board_length
    ):
        raise ValueError(
            f"Invalid move, row or column out of bounds: {row_index}, {col_index}"
        )

    # check if the move is valid
    if game["board"][row_index][col_index] != "_":
        raise ValueError("Square already occupied")

    game["board"][row_index][col_index] = game["agent_symbol"]


def check_winner(board: list[list[str]]) -> Literal["x", "o", "draw", None]:
    board_length = len(board)
    # check rows
    for row in board:
        if row.count(row[0]) == board_length and row[0] != "_":
            return row[0]
    # check columns
    for col in range(board_length):
        if [board[row][col] for row in range(board_length)].count(
            board[0][col]
        ) == board_length and board[0][col] != "_":
            return board[0][col]

    # top right to bottom left
    upward_diagonal = [board[i][board_length - i - 1] for i in range(board_length)]
    if (
        upward_diagonal.count(upward_diagonal[0]) == board_length
        and upward_diagonal[0] != "_"
    ):
        return upward_diagonal[0]

    # top left to bottom right
    downward_diagonal = [board[i][i] for i in range(board_length)]
    if (
        downward_diagonal.count(downward_diagonal[0]) == board_length
        and downward_diagonal[0] != "_"
    ):
        return downward_diagonal[0]

    # check for draw
    if all(cell != "_" for row in board for cell in row):
        return "draw"
    return None


def get_trajectory_messages(trajectory: art.Trajectory) -> art.Messages:
    messages: art.Messages = []
    for item in trajectory.messages_and_choices:
        # if item is not a dict, convert it to a dict
        if not isinstance(item, dict):
            item = item.to_dict()

        # check if item is a choice
        if "message" in item:
            messages.append(
                {"role": "assistant", "content": item["message"]["content"]}
            )
        else:
            # otherwise it's a message
            messages.append(item)
    return messages



import asyncio
from functools import wraps
from typing import Callable, Optional


def limit_concurrency(n: int, derive_key: Optional[Callable[..., str]] = None):
    semaphores = {}

    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            if derive_key:
                key = derive_key(*args, **kwargs)
            else:
                key = "default"

            if key not in semaphores:
                semaphores[key] = asyncio.Semaphore(n)

            async with semaphores[key]:
                return await func(*args, **kwargs)

        return wrapper

    return decorator
